'------------------------------------------------------------------------------
' 
' 	BSD License
'   -----------
' 
'   Copyright (c) 2012 Elba Corp.  All rights reserved.
' 
'      Elba Corp.
'      10915 SW Avocet Ct.
'      Beaverton, OR  97007
' 
'   Redistribution and use in source and binary forms, with or without
'   modification, are permitted only for use by licensees of the ZBasic
'   Compiler for generic AVR devices and only then if all the following
'   conditions are met:
' 
'   - Redistributions of source code must retain the above copyright notice,
'     this list of conditions and the following disclaimer.
' 
'   - Redistributions in binary form must reproduce the above copyright notice,
'     this list of conditions and the following disclaimer in the documentation
'     and/or other materials provided with the distribution.
' 
'   - Neither the name of the copyright owner nor the names of its contributors
'     may be used to endorse or promote products derived from this software
'     without specific prior written permission of the copyright owner.
' 
'   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
'   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
'   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
'   ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
'   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
'   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
'   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
'   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
'   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
'   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
'   POSSIBILITY OF SUCH DAMAGE.
' 
'------------------------------------------------------------------------------

'------------------------------------------------------------------------------
' 
' The code in this module presents a simple demonstration of how to utilize
' some of the functionality implemented by FatFS.  It is not a complete
' demonstration by any means but it should provide a good starting point for
' using FatFS in a ZBasic application.
' 
'------------------------------------------------------------------------------

' Declare the test monitor entry point in FatFS_test.c.
Declare Sub FatFS_test()

' These two data elements are used to output file attributes.  Each
' character in the fileAttrChar string corresponds to the attribute
' mask bit in the same position of the fileAttrMask array.
Const fileAttrChar as String = "DRHSA"
Dim fileAttrMask as ByteVectorData ({ AM_DIR, AM_RDO, AM_HID, AM_SYS, AM_ARC })

Dim fs as FATFS

Sub Main()
	Dim s as String

	' initialize the external timekeeping chip, set ZBasic time
	Call SetZBasicRTC()

   ' perform hardware initialization
   Call FatFS_init()

	' initialize the SD card parameters
	Dim b as Byte
	b = disk_initialize(0)
#if 0
	' Call the test monitor (never returns).  Note that echo needs to
	' be turned off if using the IDE or a terminal emulator that also echos.
	Register.Console.Echo = False
	Call FatFS_test()
#else
	' mount drive 0
	Dim res as FRESULT
	res = f_mount(0, fs)

	Call TestWrite()
	Call DisplayDir("/")
#endif
End Sub

'----------------------------------------------------------------------------
'
' This routine creates a new file (or overwrites an existing file) and writes
' some data to it.
'
Sub TestWrite()
	Dim res as FRESULT
	Dim f as FIL
	Dim s as String

	' create the file, overwriting an existing file
	s = NullTermString("/thefox.txt")
	res = f_open(f, StrPtr(s), FA_CREATE_ALWAYS Or FA_WRITE)
	If (res <> FR_OK) Then
		Debug.Print "f_open() returns "; res
		Exit Sub
	End If

	' write some data to the file
	s = NullTermString("The quick brown fox jumped over the lazy dog." & Chr(&H0d) & Chr(&H0a))
	Dim dataLen as UnsignedInteger
	Dim writeCnt as UnsignedInteger
	dataLen = CUInt(Len(s)) - 1
	res = f_write(f, StrPtr(s), dataLen, writeCnt)
	If (res <> FR_OK) Then
		Debug.Print "f_write() of "; dataLen; " bytes returns "; res
	ElseIf (writeCnt <> dataLen) Then
		Debug.Print "f_write() reports "; writeCnt; " of "; dataLen; " bytes written"
	End If

	' close the file
	res = f_close(f)
End Sub

'----------------------------------------------------------------------------
'
' This routine displays the content of the named directory.
'
Sub DisplayDir(ByVal dirName as String)
	Dim res as FRESULT

	' display a directory listing of the root
	Dim d as DIR
	Dim s as String
	s = NullTermString(dirName)
	res = f_opendir(d, StrPtr(s))
	If (res = FR_OK) Then
		Dim fileSize as UnsignedLong = 0
		Dim dirCnt as UnsignedInteger = 0
		Dim fileCnt as UnsignedInteger = 0
		Dim finfo as FILINFO
		Do
			' read the next directory entry
			res = f_readdir(d, finfo)
			If ((res <> FR_OK) Or (finfo.fname(1) = 0)) Then
				Exit Do
			End If
			If (CBool(finfo.fattrib And AM_DIR)) Then
				dirCnt = dirCnt + 1
			Else
				fileCnt = fileCnt + 1
				fileSize = fileSize + finfo.fsize
			End If
			
			' output the attribute indicators
			Dim i as Integer
			For i = 1 to SizeOf(fileAttrMask)
				Debug.Print IIF(CBool(finfo.fattrib And fileAttrMask(i)), Chr(Asc(fileAttrChar, i)), "-");
			Next i
			Debug.Print " ";
			
			' output the file date/time
			Debug.Print Shr(finfo.fdate, 9) + 1980;
			Call outputDigits(CByte(Shr(finfo.fdate, 5) And &H0f), "/")
			Call outputDigits(CByte(finfo.fdate And &H1f), "/")
			Call outputDigits(CByte(Shr(finfo.ftime, 11) And &H1f), " ")
			Call outputDigits(CByte(Shr(finfo.ftime, 5) And &H3f), ":")
			Call outputDigits(Shl(CByte(finfo.ftime And &H1f), 1), ":")

			' output the file size
			Debug.Print Right("          " & CStr(finfo.fsize), 11); "  ";

			' output the file name (N.B.: fname array is null-terminated)
			Debug.Print finfo.fname
		Loop
		
		' output disk statistics
		Debug.Print fileCnt; " files, "; dirCnt; " directories, "; fileSize; " bytes total"
		Dim freeClusters as UnsignedLong
		res = getFreeClusters(StrPtr(s), freeClusters, fs)
		If (res = FR_OK) Then
			Debug.Print Shr(freeClusters * CULng(fs.csize), 1); "K bytes free"
		End If
	End If
End Sub

'----------------------------------------------------------------------------
'
'' getFreeClusters
'
' This helper function needs to be written in C because the FatFS function f_getfree()
' takes a pointer-to-a-pointer parameter that isn't supported by ZBasic.
'
Declare Function getFreeClusters(ByRef dir() as Byte, ByRef freeClusters as UnsignedLong, ByRef fs as FATFS) as FRESULT
#c
FRESULT
getFreeClusters(uint8_t *dir, uint32_t *freeClusters, FATFS *fs)
{
	return(f_getfree((TCHAR *)dir, freeClusters, &fs));
}
#endc

'------------------------------------------------------------------------------
'
'' outputDigits
'
' Output two digits corresponding to a passed value with an optional separator.
'
Private Sub outputDigits(ByVal val as Byte, ByVal sep as String = "")
	Dim digit1 as Byte
	Dim digit2 as Byte
	digit2 = (val Mod 10)
	val = val \ 10
	digit1 = (val Mod 10)
	Debug.Print sep; digit1; digit2;
End Sub

'----------------------------------------------------------------------------
'
'' StrPtr
'
' This helper function is used to convert the address where a string's characters
' are stored into a pointer-to-a-byte as needed by some of the FatFS functions.
' Note that FatFS functions expect a string to be in RAM - a circumstance that
' may or may not be true for an arbitrary string in ZBasic.  You can force a
' string to reside in RAM by performing an operation on it like Left(), Right(),
' Mid(), etc.  Alternatively, the NullTermString() function below may be used
' to (a) ensure that the string resides in RAM and (b) that it is null-terminated
' as C functions expect.
'
Function StrPtr(ByRef st as String) as Byte ByRef
	StrPtr.DataAddress = StrAddress(st)
End Function

'----------------------------------------------------------------------------
'
'' NullTermString
'
' Given a string (of any length less than the maximum string length)
' and located in RAM, Program Memory or Persistent Memory, return a
' new string that is guaranteed to reside in RAM and is null terminated.
'
Function NullTermString(ByVal st as String) as String
    Const MaxStrLen as Integer = 255
    Dim slen as Integer
    NullTermString = ""
    slen = Len(st)
    If ((slen = 0) Or (Asc(st, slen) <> 0)) Then
        ' zero-length string or string not already null terminated
        If (slen < 2) Then
            ' resulting string will be 1 or 2 characters, construct in-situ
            Dim strData() as Byte Based NullTermString.DataAddress
            strData(1) = CByte(slen + 1)
            strData(2) = &He5
            strData(3) = IIF(slen = 0, 0, Asc(st, 1))
            strData(4) = 0
        ElseIf (slen < MaxStrLen) Then
            ' resulting string will be at least 3 characters in allocated RAM
            NullTermString = st & Chr(0)
        End If
    Else
        ' string is already null-terminated, ensure it is in RAM
        Select Case StrType(st)
        Case &He2, &He3
            ' source string is in ProgMem or Persistent Memory
            NullTermString = Left(st, slen)
        Case Else
            ' source string is already in RAM
            NullTermString = st
        End Select
    End If
End Function
