'------------------------------------------------------------------------------
' 
' 	BSD License
'   -----------
' 
'   Copyright (c) 2012 Elba Corp.  All rights reserved.
' 
'      Elba Corp.
'      10915 SW Avocet Ct.
'      Beaverton, OR  97007
' 
'   Redistribution and use in source and binary forms, with or without
'   modification, are permitted only for use by licensees of the ZBasic
'   Compiler for generic AVR devices and only then if all the following
'   conditions are met:
' 
'   - Redistributions of source code must retain the above copyright notice,
'     this list of conditions and the following disclaimer.
' 
'   - Redistributions in binary form must reproduce the above copyright notice,
'     this list of conditions and the following disclaimer in the documentation
'     and/or other materials provided with the distribution.
' 
'   - Neither the name of the copyright owner nor the names of its contributors
'     may be used to endorse or promote products derived from this software
'     without specific prior written permission of the copyright owner.
' 
'   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
'   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
'   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
'   ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
'   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
'   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
'   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
'   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
'   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
'   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
'   POSSIBILITY OF SUCH DAMAGE.
' 
'------------------------------------------------------------------------------

'----------------------------------------------------------------------------
'
' The routines in this file are based on the FatFS AVR sample file mmc.c.
'
'----------------------------------------------------------------------------

' import the necessary header files
#import Public "FatFS/diskio.h" "#all-define"
#import Public "FatFS/ff.h" "#all-define"

' configuration parameters
Private Const chan as Byte = 1
Private Const csPin as Byte = A.6  ' change this to match your setup

' These timer variables are decremented by the compare match ISR.
Private Timer0 as UnsignedInteger
Private Timer1 as Byte Attribute(Volatile)
Private Timer2 as Byte Attribute(Volatile)

Private stat as Byte
Private cardType as Byte

'----------------------------------------------------------------------------
' definitions for MMC/SDC command codes
Private Const CMD0 as Byte = 0                     ' GO_IDLE_STATE
Private Const CMD1 as Byte = 1                     ' SEND_OP_COND as Byte = MMC
Private Const ACMD41 as Byte = &H80+41             ' SEND_OP_COND as Byte = SDC
Private Const CMD8 as Byte = 8                     ' SEND_IF_COND
Private Const CMD9 as Byte = 9                     ' SEND_CSD
Private Const CMD10 as Byte = 10                   ' SEND_CID
Private Const CMD12 as Byte = 12                   ' STOP_TRANSMISSION
Private Const ACMD13 as Byte = &H80+13             ' SD_STATUS as Byte = SDC
Private Const CMD16 as Byte = 16                   ' SET_BLOCKLEN
Private Const CMD17 as Byte = 17                   ' READ_SINGLE_BLOCK
Private Const CMD18 as Byte = 18                   ' READ_MULTIPLE_BLOCK
Private Const CMD23 as Byte = 23                   ' SET_BLOCK_COUNT as Byte = MMC
Private Const ACMD23 as Byte = &H80+23             ' SET_WR_BLK_ERASE_COUNT as Byte = SDC
Private Const CMD24 as Byte = 24                   ' WRITE_BLOCK
Private Const CMD25 as Byte = 25                   ' WRITE_MULTIPLE_BLOCK
Private Const CMD55 as Byte = 55                   ' APP_CMD
Private Const CMD58 as Byte = 58                   ' READ_OCR

' definitions for card type flags
Private Const CT_MMC as Byte = &H01                ' MMC ver 3
Private Const CT_SD1 as Byte = &H02                ' SD ver 1
Private Const CT_SD2 as Byte = &H04                ' SD ver 2
Private Const CT_SDC as Byte = (CT_SD1 Or CT_SD2)  ' SD
Private Const CT_BLOCK as Byte = &H08              ' Block addressing

'----------------------------------------------------------------------------
'
' N.B.: This ISR is written for the ZX devices based on the
'       mega644P and xmega chips.  Changes may be required for
'       other devices and chips operating at other frequencies.
' 
' mega644P:
' This ISR is called on a Timer0 CompareB match and is set up
' to decrement timer variables at a rate of 100Hz.  Timer0 is used
' for the ZBasic RTC and runs at a rate that is dependent on the main
' clock frequency, the RTC frequency and the RTC scale factor.  On a
' ZX-24n the main clock frequency is 14.7MHz and the TCNT0 register
' changes at a rate of 230.4KHz.  This can be divided down to 100Hz
' by dividing by 2304.  However, since the range of the count register
' of Timer0 is 0 to 224 this can't be achieved directly but we observe
' that 144 * 16 equals 2304 so if we can arrange to get an interrupt
' every 144 counts we can easily count interrupts 16 in the ISR to
' realize the 100Hz rate (10mS).
'
' In order to get an interrupt every 160mS (16 * 10mS) we need to
' change the CompareB register on each interrupt, advancing it by
' 144 but limiting the value to no more than 224 (since the counter
' value will never exceed 224).
'
' Note that the X-10 low level functions also use the CompareB match
' register - that use is mutually exclusive with this use.  Also, this
' code cannot be used on AVR devices that have only one compare match
' register on the RTC timer.  Examples of such devices are the mega8,
' mega16, mega32, mega64 and mega128.  See the description of OpenX10()
' for a list of devices that are incompatible both with this code and
' the low-level X-10 functionality.
'
' xmega devices:
' For xmega the idea is the same as described above but the RTC timer
' is a 16-bit timer that changes at a rate of 115.2KHz.  All xmega
' devices use the CCA compare register of TimerC1 for the RTC.  The
' X-10 low level functions use the CCB compare register so use of those
' is mutually exclusive with using CCB for generating the 100Hz timing.
'
' To adapt this ISR for other target devices you must first confirm
' that the RTC timer has a second compare register and the rate at which
' the timer's count register changes (this is referred to as F_RTC_TIMER
' in the  ZBasic System Library manual section 3).  Then, determine the
' divisor necessary to yield a 100Hz rate (F_RTC_TIMER / 100).  If the
' resulting divisor is greater than that used in the CompareA register
' further division in the ISR as described above will be required.
'
#if (Option.TimerRTC = "Timer0")
ISR TIMER0_COMPB()
	Const CountMask as Byte = &H0f
	Const Delta100Hz as Byte = 144
#elseif (Option.TimerRTC = "Timer2")
ISR TIMER2_COMPB()
	Const CountMask as Byte = &H0f
	Const Delta100Hz as Byte = 144
#elseif (Option.TimerRTC = "TimerC1")
ISR TCC1_CCB()
	Const CountMask as Byte = &H07
	Const Delta100Hz as UnsignedInteger = 144
#else
ISR FIXME()
#endif
	' advance the interrupt counter
	Static count as Byte
	count = (count + 1) And CountMask
	If (count = 0) Then
		' a 10mS interval has passed
		Timer0 = Timer0 + 1
		If (Timer1 <> 0) Then
			Timer1 = Timer1 - 1
		End If
		If (Timer2 <> 0) Then
			Timer2 = Timer2 - 1
		End If
	End If

	' advance the compare register for the next interrupt
#if (Option.TimerRTC = "Timer0")
	Dim delta as Byte
	delta = Register.OCR0A - Register.OCR0B
	If (delta >= Delta100Hz) Then
		Register.OCR0B = Register.OCR0B + Delta100Hz
	Else
		Register.OCR0B = Delta100Hz - delta
	End If
#elseif (Option.TimerRTC = "Timer2")
	Dim delta as Byte
	delta = Register.OCR2A - Register.OCR2B
	If (delta >= Delta100Hz) Then
		Register.OCR2B = Register.OCR2B + Delta100Hz
	Else
		Register.OCR2B = Delta100Hz - delta
	End If
#elseif (Option.TimerRTC = "TimerC1")
	Dim delta as UnsignedInteger
	delta = Register.TCC1_CCA - Register.TCC1_CCB
	If (delta >= Delta100Hz) Then
		Register.TCC1_CCB = Register.TCC1_CCB + Delta100Hz
	Else
		Register.TCC1_CCB = Delta100Hz - delta
	End If
#endif
End ISR

'----------------------------------------------------------------------------
'
' Read and optionally reset the 16-bit timer (used only by the test "monitor").
'
Public Function getTimer(ByVal reset as Boolean) as UnsignedInteger Attribute(Used, "Alias:timer")
	Atomic
		getTimer = Timer0
		If (reset) Then
			Timer0 = 0
		End If
	End Atomic
End Function

'----------------------------------------------------------------------------
'
' Read a character from the serial channel.
'
Public Function uart_get() as Byte Attribute(Used, "Alias:uart_get")
	uart_get = Console.Read()
End Function

'----------------------------------------------------------------------------
'
' Send a character to the serial channel.
'
Public Sub uart_put(ByVal c as Byte) Attribute(Used, "Alias:uart_put")
	If (c = &H0a) Then
		Debug.Print
	ElseIf (c <> &H0d) Then
		Debug.Print Chr(c);
	End If
End Sub

'----------------------------------------------------------------------------
'
'' FatFS_init
'
' Perform initialization of the hardware related to MMC FatFS.
'
Public Sub FatFS_init()
#if (Option.TimerRTC = "Timer0")
	' enable interrupts on Timer0 CompareB match
	Register.TIFR0 = &H04
	Call SetBits(Register.TIMSK0.DataAddress, &H04, &H04)
#elseif (Option.TimerRTC = "Timer2")
	' enable interrupts on Timer2 CompareB match
	Register.TIFR2 = &H04
	Call SetBits(Register.TIMSK2.DataAddress, &H04, &H04)
#elseif (Option.TimerRTC = "TimerC1")	' xmega
	' enable interrupts on TimerC1 CompareB match
	Register.TCC1_INTFLAGS = &H20
	Call SetBits(Register.TCC1_INTCTRLB.DataAddress, &H0C, &H08)
#else
	#error Need initialization code
#endif

	' set the disk status to "uninitialized"
	stat = STA_NOINIT
End Sub

'----------------------------------------------------------------------------
'
' Initialize the SD/MMC Disk
'
Public Function disk_initialize(ByVal drv as Byte) as Byte Attribute(Used, "Alias:disk_initialize")
	Dim i as Byte
	Dim res as Byte
	Dim cmd as Byte
	Dim ty as Byte
	Dim ocr(1 to 4) as Byte

	stat = STA_NOINIT
	disk_initialize = stat
	If CBool(drv) Then
		Exit Function
	End If

	' Open the SPI channel.  Initially, the clock speed must be < 400KHz.
	Call OpenSPI(chan, &H03, csPin)
	Call SPIStart(chan, &H03)

	' send 80 SPI clock cycles to put the card in SPI mode
	For i = 1 to 10
		res = SPIGetByte(chan, &Hff)
	Next i

	' attempt to determine the card type
	cardType = 0
	res = send_cmd(CMD0, 0)
	If (res = 1) Then
		' Enter Idle State
		Timer1 = 100						' Initialization timeout of 1000 msec
		res = send_cmd(CMD8, &H1AA)
		If (res = 1) Then
			res = CByte(SPIGetData(chan, 4, ocr, &Hff))
			If ((res = 4) And (ocr(3) = &H01) And (ocr(4) = &Haa)) Then
				' Wait for leaving idle state (ACMD41 with HCS bit)
				Do While ((Timer1 > 0) And (send_cmd(ACMD41, Shl(1, 30)) > 0))
				Loop
				' Check CCS bit in the OCR
				If ((Timer1 > 0) And (send_cmd(CMD58, 0) = 0)) Then
					res = CByte(SPIGetData(chan, 4, ocr, &Hff))
					If ((res = 4) And CBool(ocr(1) And &H40)) Then
						cardType = CT_SD2 Or CT_BLOCK
					Else
						cardType = CT_SD2
					End If
				End If
			End If
		Else	'SDv1 or MMCv3
			res = send_cmd(ACMD41, 0)
			If (res <= 1) Then
				cardType = CT_SD1
				cmd = ACMD41
			Else
				cardType = CT_MMC
				cmd = CMD1
			End If
			Do While ((Timer1 > 0) And (send_cmd(cmd, 0) > 0))
			Loop
			If ((Timer1 > 0) Or (send_cmd(CMD16, 512) <> 0)) Then
				cardType = 0
			End If
		End If
	End If
	
	If (cardType <> 0) Then
		stat = stat And Not STA_NOINIT
		Call OpenSPI(chan, &H80, csPin)	' switch to high speed SPI
	End If
	disk_initialize = stat
End Function

'----------------------------------------------------------------------------
'
' Get Disk Status
'
Public Function disk_status(ByVal drv as Byte) as Byte Attribute(Used, "Alias:disk_status")
	If (drv <> 0) Then
		disk_status = STA_NOINIT
	Else
		disk_status = stat
	End If
End Function

'----------------------------------------------------------------------------
'
' Read Sector(s)
'
Public Function disk_read(ByVal drv as Byte, ByRef buf() as Byte, _
		ByVal sector as UnsignedLong, ByVal count as Byte) as DRESULT _
		Attribute(Used, "Alias:=DRESULT disk_read(uint8_t, uint8_t *, uint32_t, uint8_t)")
	' check parameters and disk status
	If ((drv <> 0) Or (count = 0)) Then
		disk_read = RES_PARERR
		Exit Function
	End If
	If (CBool(stat And STA_NOINIT)) Then
		disk_read = RES_NOTRDY
		Exit Function
	End If
	disk_read = RES_ERROR
	
	' convert to byte address if needed
	If ((cardType And CT_BLOCK) = 0) Then
		sector = sector * 512
	End If
	
	If (count = 1) Then
		' single block read
		If ((send_cmd(CMD17, sector) = 0) And (read_data(buf.DataAddress, 512) <> 0)) Then
			disk_read = RES_OK
		End If
	Else
		' multiple block read
		If (send_cmd(CMD18, sector) = 0) Then
			Dim bufAddr as UnsignedInteger
			bufAddr = buf.DataAddress
			Do
				If (read_data(bufAddr, 512) = 0) Then
					Exit Do
				End If

				bufAddr = bufAddr + 512
				count = count - 1
			Loop While (count > 0)
			If (count = 0) Then
				disk_read = RES_OK
			End If
			' stop transmission
			Dim res as Byte
			res = send_cmd(CMD12, 0)
		End If
	End If
	Call deselect_card()
End Function

'----------------------------------------------------------------------------
'
' Write Sector(s)
'
Public Function disk_write(ByVal drv as Byte, ByVal buf() as Byte, _
		ByVal sector as UnsignedLong, ByVal count as Byte) as DRESULT _
		Attribute(Used, "Alias:=DRESULT disk_write(uint8_t, const uint8_t *, uint32_t, uint8_t)")
	If ((drv <> 0) Or (count = 0)) Then
		disk_write = RES_PARERR
		Exit Function
	End If
	If (CBool(stat And STA_NOINIT)) Then
		disk_write = RES_NOTRDY
		Exit Function
	End If
	If (CBool(stat And STA_PROTECT)) Then
		disk_write = RES_WRPRT
		Exit Function
	End If
	disk_write = RES_ERROR

	' convert to byte address if needed
	If ((cardType And CT_BLOCK) = 0) Then
		sector = sector * 512
	End If

	If (count = 1) Then
		' single block write
		If ((send_cmd(CMD24, sector) = 0) And (write_data(buf.DataAddress, &Hfe) <> 0)) Then
			disk_write = RES_OK
		End If
	Else
		' multiple block write
		Dim res as Byte
		If CBool(cardType And CT_SDC) Then
			res = send_cmd(ACMD23, CULng(count))
		End If
		If (send_cmd(CMD25, sector) = 0) Then
			Dim bufAddr as UnsignedInteger
			bufAddr = buf.DataAddress
			Do
				If (write_data(bufAddr, &Hfc) = 0) Then
					Exit Do
				End If

				bufAddr = bufAddr + 512
				count = count - 1
			Loop While (count > 0)
			If (count = 0) Then
				disk_write = RES_OK
			End If

			' stop transmission
			res = send_cmd(CMD12, 0)
		End If
	End If
	Call deselect_card()
End Function

'----------------------------------------------------------------------------
'
' Perform low-level control operations.
'
Public Function disk_ioctl(ByVal drv as Byte, ByVal ctrl as Byte, ByRef data() as Byte) as DRESULT _
		Attribute(Used, "Alias:=DRESULT disk_ioctl(uint8_t, uint8_t, void *)")
	If (drv <> 0) Then
		disk_ioctl = RES_PARERR
		Exit Function
	End If
	disk_ioctl = RES_ERROR
	
	If (ctrl = CTRL_POWER) Then
		' not implemented
	Else
		Dim res as Byte
		Dim csd(1 to 16) as Byte
		If CBool(Stat And STA_NOINIT) Then
			disk_ioctl = RES_NOTRDY
			Exit Function
		ElseIf (ctrl = CTRL_SYNC) Then
			' Make sure that no pending write process. Do not remove this or written sector might not left updated.
			If (select_card() <> 0) Then
				disk_ioctl = RES_OK
			End If
		ElseIf (ctrl = GET_SECTOR_COUNT) Then
			' Get number of sectors on the disk (DWORD)
			If (send_cmd(CMD9, 0) = 0) And (read_data(csd.DataAddress, 16) = 1) Then
				Dim dval as UnsignedLong Based data.DataAddress
				Dim csize as UnsignedInteger
				If ((csd(1) And &Hc0) = &H40) Then
					' SDC ver 2.00
					csize = MakeWord(csd(10), csd(9)) + 1
					dval = Shl(CULng(csize), 10)
				Else
					' SDC ver 1.XX or MMC
					Dim n as Byte
					n = (csd(6) And &H0f) + Shr(csd(11) And &H80, 7) + Shl(csd(10) And &H03, 1) + 2
					csize = CUInt(Shr(csd(9), 6)) + Shl(CUInt(csd(8)), 2) + Shl(CUInt(csd(7) And 3), 10) + 1
					dval = Shl(CULng(csize), n - 9)
				End If
				disk_ioctl = RES_OK
			End If
		ElseIf (ctrl = GET_SECTOR_SIZE) Then
			' Get R/W sector size (WORD)
			Dim wval as UnsignedInteger Based data.DataAddress
			wval = 512
			disk_ioctl = RES_OK
		ElseIf (ctrl = GET_BLOCK_SIZE) Then
			' Get erase block size in unit of sector (DWORD)
			Dim dval as UnsignedLong Based data.DataAddress
			If (CBool(cardType And CT_SD2)) Then
				' SDv2
				If (send_cmd(ACMD13, 0) = 0) Then
					res = SPIGetByte(chan, &Hff)
					If (read_data(csd.DataAddress, 16) = 1) Then
						Dim i as Byte
						For i = 1 to 64 - 16
							' read the remaining data bytes
							res = SPIGetByte(chan, &Hff)
						Next i
						dval = Shl(16, Shr(csd(11), 4))
						disk_ioctl = RES_OK
					End If
				End If
			Else
				' SDv1 or MMCv3
				If (send_cmd(CMD9, 0) = 0) And (read_data(data.DataAddress, 16) = 1) Then
					If (CBool(cardType And CT_SD1)) Then
						' SDv1
						dval = CULng(Shl(csd(11) And &H3f, 1)) + Shl(CULng(Shr(csd(12) And &H80, 7) + 1), Shr(csd(14), 6) - 1)
					Else
						' MMCv3
						dval = CULng(Shl(CUInt(csd(11)) And &H7c, 2) + 1) * CULng(Shl(csd(12) And &H03, 3) + Shr(csd(12) And &H7c, 5) + 1)
					End If
					disk_ioctl = RES_OK
				End If
			End If
		ElseIf (ctrl = MMC_GET_TYPE) Then
			' Get card type flags (1 byte)
			data(1) = cardType
			disk_ioctl = RES_OK
		ElseIf (ctrl = MMC_GET_CSD) Then
			' Receive CSD as a data block (16 bytes)
			If (send_cmd(CMD9, 0) = 0) And (read_data(data.DataAddress, 16) = 1) Then
				disk_ioctl = RES_OK
			End If
		ElseIf (ctrl = MMC_GET_CID) Then
			' Receive CID as a data block (16 bytes)
			If (send_cmd(CMD10, 0) = 0) And (read_data(data.DataAddress, 16) = 1) Then
				disk_ioctl = RES_OK
			End If
		ElseIf (ctrl = MMC_GET_OCR) Then
			' Receive OCR as an R3 resp (4 bytes)
			If (send_cmd(CMD58, 0) = 0) And (SPIGetData(chan, 4, data, &Hff) = 4) Then
				disk_ioctl = RES_OK
			End If
		ElseIf (ctrl = MMC_GET_SDSTAT) Then
			' Receive SD status as a data block (64 bytes)
			If (send_cmd(ACMD13, 0) = 0) Then
				res = SPIGetByte(chan, &Hff)
				If (read_data(data.DataAddress, 64) = 1) Then
					disk_ioctl = RES_OK
				End If
			End If
		Else
		   disk_ioctl = RES_PARERR
		End If
		Call deselect_card()
	End If
End Function

'----------------------------------------------------------------------------
'
' This is a real time clock service called from the FatFs module. It is needed
' only if FatFS is configured for write operations.  A valid time must be
' returned even if the system does not support a real time clock.       
'
' The time value to be returned has sub-fields as follows:
'    value           bits
'    --------------  -----
'    year - 1980     31-25
'    month (1-12)    24-21
'    day (1-31)      20-16
'    hour (0-23)     15-11
'    minute (0-59)   10-5
'    2-second (0-29) 4-0
'
' Note that the "packed" forms of ZBasic time and date are nearly the same
' as the required format.  One difference is that in ZBasic packed
' date form the year is relative to 1999 rather than being relative to 1980.
' The high word of the packed time value exactly matches the required format
' and the low word is not used.
'
Public Function get_fattime() as UnsignedLong Attribute(Used, "Alias:get_fattime")
#if Option.RTCEnabled
	Dim time as UnsignedLong
	Dim date as UnsignedInteger
	Atomic
		' retrieve ZBasic time and date in packed form
		time = GetTimeValue()
		date = GetDateValue()
	End Atomic

	' combine the date/time values, correct for different year 0
	get_fattime = MakeDword(HiWord(time), date + Shl(1999 - 1980, 9))
#else
	' no RTC, return a non-zero value (add code here to read from external RTC)
	get_fattime = 1
#endif
End Function

'----------------------------------------------------------------------------
'
' Wait for the SD card to become ready.  Return 1 if OK, 0 on timeout.
'
Private Function wait_ready() as Byte
	wait_ready = 0
	Timer2 = 50		' set the timeout for 500ms
	Do
		If (SPIGetByte(chan, &Hff) = &Hff) Then
			wait_ready = 1
		End If
	Loop While ((wait_ready = 0) And (Timer2 > 0))
End Function

'----------------------------------------------------------------------------
'
' Deselect the card and release SPI bus
'
Private Sub deselect_card()
	' deassert the chip select and then apply 8 clocks
	Call SPIStop(chan, &H02)
	Dim b as Byte
	b = SPIGetByte(chan, &Hff)
End Sub

'----------------------------------------------------------------------------
'
' Select the card and wait for ready
'
Private Function select_card() as Byte
	' assert chip select and apply 8 clocks with MOSI high
	Call SPIStart(chan, &H05)
	Dim b as Byte
	b = SPIGetByte(chan, &Hff)
	
	' wait for the card to be ready
	If (wait_ready() <> 0) Then
		select_card = 1
	Else
		Call deselect_card()
		select_card = 0
	End If
End Function

'----------------------------------------------------------------------------
'
' Send a command packet to MMC
'
Private Function send_cmd(ByVal cmd as Byte, ByVal arg as UnsignedLong) as Byte
	Dim res as Byte

	' ACMD<n> is the command sequense of CMD55-CMD<n>
	If CBool(cmd And &H80) Then
		res = send_cmd(CMD55, 0)
		If (res > 1) Then
			send_cmd = res
			Exit Function
		End If
		cmd = cmd And &H7F
	End If

	' Select the card and wait for ready
	Call deselect_card()
	If (select_card() = 0) Then
		send_cmd = &HFF
		Exit Function
	End If

	' Send command packet
	Dim buf(1 to 6) as Byte
	Dim data() as Byte Based arg.DataAddress
	buf(1) = cmd Or &H40				' Start + Command index
	buf(2) = HiByte(HiWord(arg))		' Argument[31..24]
	buf(3) = LoByte(HiWord(arg))		' Argument[23..16]
	buf(4) = HiByte(CUInt(arg))			' Argument[15..8]
	buf(5) = LoByte(CUInt(arg))			' Argument[7..0]
	If (cmd = CMD0) Then
		buf(6) = &H95					' Valid CRC for CMD0(0)
	ElseIf (cmd = CMD8) Then
		buf(6) = &H87					' Valid CRC for CMD8(0x1AA)
	Else
		buf(6) = &H01					' Dummy CRC + Stop
	End If
	res = CByte(SPIPutData(chan, 6, buf))

	' Receive command response
	Dim n as Byte
	If (cmd = CMD12) Then
		n = SPIGetByte(chan, &Hff)		' Skip a stuff byte when stop reading
	End If
	n = 10								' Wait for a valid response in timeout of 10 attempts
	Do
		send_cmd = SPIGetByte(chan, &Hff)
		n = n - 1
	Loop While (CBool(send_cmd And &H80) And (n > 0))
End Function

'----------------------------------------------------------------------------
'
' Receive a data packet from MMC
'
Private Function read_data(ByVal bufAddr as UnsignedInteger, ByVal count as UnsignedInteger) as Byte
	read_data = 0

	' wait for a packet timeout
	Timer1 = 20
	Dim token as Byte
	Do
		token = SPIGetByte(chan, &Hff)
	Loop While ((token = &Hff) And (Timer1 > 0))
	If (token = &Hfe) Then
		' read the data, discard the CRC
		Dim res as Byte
		Dim buf() as Byte Based bufAddr
		res = CByte(SPIGetData(chan, count, buf, &Hff))
		res = SPIGetByte(chan, &Hff)
		res = SPIGetByte(chan, &Hff)
		read_data = 1
	End If
End Function

'----------------------------------------------------------------------------
'
' Send a data packet to MMC
'
Private Function write_data(ByVal bufAddr as UnsignedInteger, ByVal token as Byte) as Byte
	Dim res as Byte
	
	write_data = 1
	If (wait_ready() <> 0) Then
		' send the data token
		res = SPIPutByte(chan, token)
		
		If (token <> &Hfd) Then
			' send the data and a dummy CRC value
			Dim buf() as Byte Based bufAddr
			res = CByte(SPIPutData(chan, 512, buf))
			res = SPIPutByte(chan, &Hff)
			res = SPIPutByte(chan, &Hff)
			res = SPIGetByte(chan, &Hff)
			If ((res And &H1f) <> &H05) Then
				write_data = 0
			End If
		End If
	Else
		write_data = 0
	End If
End Function



